/* 
 * firewall.c: iptables rules parser
 * Copyright (C) 2001 Axel Burri <aburri@ee.ethz.ch>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */


/*
 * EBNF for input file:
 *
 * line:       {("!!" inline) | ("!" inline) | ("#" comment) | f_command | io_command}
 *
 * f_command:  interface + interface + xmap args
 * io_command: chain     + interface + xmap args
 * args:       proto [options] address address [port]
 * proto:      "TCP" | "UDP" | ...        (not parsed!)
 * option:     "--" type value            (e.g. --icmp-type ping)
 * port:       string
 * xmap:       "." | "x" | "X"            (no_xmap | unidirectional | bidirectional)
 * interface:  "0".."9" | "a"             (eth0..eth9 | all)
 * chain:      "i" | "o"                  (INPUT | OUTPUT)
 * address:    string
 * inline:     {string}
 *
 */


#include <stdio.h>
#include <unistd.h>

char* shell_script_head = "#!/bin/sh\n";
char* iptables_bin = "/prodlocal/bin/iptables ";
char* default_target = "ACCEPT";

#define LINEBUF_LEN 1024
#define WORD_MAXLEN 64

#define CMDBUF_LEN 1024



#define CHAIN_INPUT 0
#define CHAIN_OUTPUT 1
#define CHAIN_FORWARD 2

#define IF_ALL -1

#define XMAP_NONE 0
#define XMAP_UNI 1
#define XMAP_BI 2



char next_word[WORD_MAXLEN] = "";

int print_comments = 0; // print commented out statements
int shell_script = 0; // generate shell-script



typedef struct {
    int chain;
    int if_in;  // in-interface
    int if_out; // out-interface
    int xmap;   // X-mapping command
    char proto[WORD_MAXLEN];
    char options[WORD_MAXLEN * 2 + 1];
    char src[WORD_MAXLEN];
    char dst[WORD_MAXLEN];
    char port[WORD_MAXLEN];
} cmd_struct;
typedef cmd_struct *cmd_ptr;

typedef struct {
    char *chain;
    char *proto;
    char *options;
    int if_in;  // in-interface
    int if_out; // out-interface
    int syn;
    char *src;
    char *sport;
    char *dst;
    char *dport;
} ipt_struct;
typedef ipt_struct *ipt_ptr;




void print_ipt_simple(FILE *out, char* options) {
    fprintf(out, "%s %s\n", iptables_bin, options);
}




void print_ipt(FILE* out, ipt_ptr ipt) {
    char opt[256] = "";
    char s[CMDBUF_LEN] = "";

    // chain
    sprintf(opt, "-A %s ", ipt->chain); strcat(s, opt);

    // proto
    sprintf(opt, "-p %s ", ipt->proto); strcat(s, opt);

    // options
    if(ipt->options) { sprintf(opt, "%s ", ipt->options); strcat(s, opt); }

    // syn
    if(!ipt->syn) strcat(s, "! --syn ");
    
    // if_in
    if(ipt->if_in != IF_ALL) { sprintf(opt, "-i eth%i ", ipt->if_in); strcat(s, opt); }

    // if_out
    if(ipt->if_out != IF_ALL) {	sprintf(opt, "-o eth%i ", ipt->if_out); strcat(s, opt); }

    // src
    sprintf(opt, "-s %s ", ipt->src); strcat(s, opt);
    if(ipt->sport) { sprintf(opt, "--sport %s ", ipt->sport); strcat(s, opt); }

    // dst
    sprintf(opt, "-d %s ", ipt->dst); strcat(s, opt);
    if(ipt->dport) { sprintf(opt, "--dport %s ", ipt->dport); strcat(s, opt); }

    // target
    sprintf(opt, "-j %s ", default_target); strcat(s, opt);

    fprintf(out, "%s %s\n", iptables_bin, s);
}



int print_command(FILE *out, cmd_ptr cmd) {
    ipt_struct ipt;
    char chain[8] = "";

    switch(cmd->chain) {
    case CHAIN_INPUT: strcpy(chain, "INPUT"); break;
    case CHAIN_OUTPUT: strcpy(chain, "OUTPUT"); break;
    case CHAIN_FORWARD: strcpy(chain, "FORWARD"); break;
    default: return -1;
    }


    ipt.chain = chain;
    ipt.proto  = cmd->proto;
    ipt.options= isempty(cmd->options) ? NULL : cmd->options;
    ipt.if_in  = cmd->if_in;
    ipt.if_out = cmd->if_out;
    ipt.syn = -1;
    ipt.src = cmd->src;
    ipt.sport = NULL;
    ipt.dst = cmd->dst;
    ipt.dport = isempty(cmd->port) ? NULL : cmd->port;

    print_ipt(out, &ipt);


    // X-mappings
    if(cmd->xmap != XMAP_NONE) {
	if(cmd->chain == CHAIN_INPUT) strcpy(chain, "OUTPUT");
	else if(cmd->chain == CHAIN_OUTPUT) strcpy(chain, "INPUT");
	ipt.if_in  = cmd->if_out;
	ipt.if_out = cmd->if_in;
	ipt.syn = (cmd->xmap == XMAP_BI);
	ipt.src = cmd->dst;
	ipt.sport = ipt.dport;
	ipt.dst = cmd->src;
	ipt.dport = NULL;
	
	print_ipt(out, &ipt);
    }

    return 0;
}




int isempty(char * s) {
    return s[0] == '\0';
}



/*
 * consume next word 
 */
void consume(char **in) {
    int i;


    *in += strspn(*in, " \t");

    i = strcspn(*in, " \t\n");

    if(i >= WORD_MAXLEN) i = 0;

    memcpy(next_word, *in, i);
    next_word[i] = '\0';
    *in += i;
}


int parse_interface(char in, int *interface) {

    switch(in) {
    case 'a': *interface = IF_ALL; break;
    case '0':
    case '1':
    case '2':
    case '3':
    case '4':
    case '5':
    case '6':
    case '7':
    case '8':
    case '9':
	*interface = in - '0'; break;
    default: return -1;
    }

    return 0;
}



int parse_xmap(char in, int *xmap) {

    switch(in) {
    case '.': *xmap = XMAP_NONE; break;
    case 'x': *xmap = XMAP_UNI; break;
    case 'X': *xmap = XMAP_BI; break;
    default: *xmap = XMAP_NONE; 
    }

    return 0;
}



int parse_options(char **in, cmd_ptr cmd) {
    int retval;
 
    cmd->options[0] = '\0';

    if(next_word[0] == '-' && next_word[1] == '-') {
	strcpy(cmd->options, next_word);	
	consume(in);
	if(isempty(next_word)) return -1;
	strcat(cmd->options, " ");
	strcat(cmd->options, next_word);
	consume(in);
    }

    return 0;
}



int parse_inline(char **in, FILE *out) {
    int retval;
 
    if(next_word[0] == '!') {
	if(next_word[1] == '!') {
	    // any command
	    fprintf(out, "%s\n", *in);
	} else {
	    // iptables inline
	    print_ipt_simple(out, *in);
	}
    }

    return 0;
}




int parse_command(char **in, cmd_ptr cmd) {
    int retval;

    if(next_word[0] == 'i')  {
	// input chain
	cmd->chain = CHAIN_INPUT;
	cmd->if_out = IF_ALL;
	// interface (in)
	if(retval = parse_interface(next_word[1], &cmd->if_in)) return retval;
	// xmap
	if(retval = parse_xmap(next_word[2], &cmd->xmap)) return retval;
    }
    else if(next_word[0] == 'o') {
	// output chain
	cmd->chain = CHAIN_OUTPUT;
	cmd->if_in = IF_ALL;
	// interface 
	if(retval = parse_interface(next_word[1], &cmd->if_out)) return retval;
	// xmap
	if(retval = parse_xmap(next_word[2], &cmd->xmap)) return retval;
    }
    else {
	cmd->chain = CHAIN_FORWARD;
	// interface (in)
	if(retval = parse_interface(next_word[0], &cmd->if_in)) return retval;
	// interface (out)
	if(retval = parse_interface(next_word[1], &cmd->if_out)) return retval;
	// xmap
	if(retval = parse_xmap(next_word[2], &cmd->xmap)) return retval;
    }
    consume(in);


    // proto
    if(isempty(next_word)) return -1;
    strcpy(cmd->proto, next_word);
    consume(in);

    // [options]
    if(isempty(next_word)) return -1;
    if((retval = parse_options(in, cmd)) < 0) return retval; 

    // address (src)
    if(isempty(next_word)) return -1;
    strcpy(cmd->src, next_word);
    consume(in);
    
    // address (dst)
    if(isempty(next_word)) return -1;
    strcpy(cmd->dst, next_word);
    consume(in);

    // port
    if(isempty(next_word)) {
	cmd->port[0] = '\0';
    } else {
	strcpy(cmd->port, next_word);
	consume(in);
    }

    return 0;
}





/*
 * parses one command (line) from input file
 */
int parse_line(char *in, FILE *out) {
    char *s = in;
    cmd_struct cmd;
    int retval = 0;


    
    consume(&s);
    if(isempty(next_word)) {
	// ignore empty lines
	return 0;
    }

    switch(next_word[0]) {
    case '!': 
	// inline command
	parse_inline(&s, out);
	break;
    case '#':
	// comment
	if(print_comments) fprintf(out, in);
	break;
    default:
	// command
	if((retval = parse_command(&s, &cmd)) < 0) return retval;
	print_command(out, &cmd);
	break;
    }
    
    return 0;
}




int parse_file(FILE *in, FILE *out, int *lines_read) {
    char line[LINEBUF_LEN];
    int retval = 0;

    *lines_read = 0;
    while(retval >= 0) {
	fgets(line, LINEBUF_LEN, in);
	if(feof(in) || !line) return 0;	
	(*lines_read)++;
	if((retval = parse_line(line, out))) return retval;
    } 

    return 0;
}




int main(int argc, char **argv)
{
    FILE *in = stdin; 
    FILE *out = stdout;
    char ch;
    char buf[100];
    int lines_read;
    int retval = 0;


   /*
    * parse arguments
    */
    while ((ch = getopt(argc, argv, "schf:")) != EOF)
    {
        switch(ch)
        {
        case 's': 
            // generate shell script
	    shell_script = -1;
            break;
        case 'c': 
            // print comments
	    print_comments = -1;
            break;
	case 'f': 
            // output file
	    if((out = fopen(optarg, "w")) == NULL) {
		    fprintf(stderr, "error opening output file.\n");
		    exit(-1);
	    }
	    break;
        default:
            fprintf(stderr, "usage: firewall-gen [-s] [-c] [-f output file] [input files...]\n");
            return (-1);
        }
    }


    /*
     * start generating output
     */
    if(shell_script) fprintf(out, shell_script_head);

    if(optind >= argc) {
	// use stdin as input
	if((retval = parse_file(in, out, &lines_read)) < 0) {
	    fprintf(stderr, "parse error: line %i\n", lines_read);
	    exit(-1);
	}
    } else {
	// use non-option arguments as input files
	while(optind < argc && retval >= 0) {
	    if((in = fopen(argv[optind++], "r")) == NULL) {
		fprintf(stderr, "error opening file '%s', exiting.\n", argv[optind - 1]);
		exit(-1);
	    }
	    if((retval = parse_file(in, out, &lines_read)) < 0) {
		fprintf(stderr, "parse error in '%s', line %i\n", argv[optind - 1], lines_read);
		exit(-1);
	    }
	}
    }



    return 0;
}

